/*
 * This is the second class and it contains the code to perform the infix 
 * expression evaluation.
 */

package main1;

import java.util.Stack;
import java.util.StringTokenizer;

/**
 * September 1, 2018
 */


class InfixParserAndCalculator {
    private static final String DELIMS = "-+*/()";

    /**
     * @param text is this expr to parse and calculate
     * @return the result after calculation
     * @throws DivideByZero thrown if division by 0 is attempted in the infix exp
     */
    static int evaluate(String text) throws DivideByZero {
        // declare the stacks.
        Stack<Integer> operandStack = new Stack<>();
        Stack<Character> operatorStack = new Stack<>();

        // Here we tokenize the expression.
        StringTokenizer st = new StringTokenizer(text, DELIMS, true);
        while(st.hasMoreTokens()) {
            // remove spaces from the token.
            String token = st.nextToken().trim();

            // if after removing spaces it is empty then ignore it.
            if(token.isEmpty()) continue;

            // if it is an operand
            //push it onto the operand stack
            if(isOperand(token)) {
                operandStack.push(Integer.parseInt(token));
            }

            //  if it is a left parenthesis
            // push it onto the operator stack
            else if("(".equals(token)) {
                operatorStack.push('(');
            }
            // evaluate the entire expression in the bracket when it closes
            else if(")".equals(token)) {
                char operatorTop = operatorStack.pop();
                while(operatorTop != '(') {
                    int b = operandStack.pop(), a =operandStack.pop();
                    int res = getResult(a,b,operatorTop);
                    operandStack.push(res);
                    operatorTop = operatorStack.pop();
                }
            }
            // evaluate considering precedence
            else if(isOperator(token)){
                while(!operatorStack.empty() && hasHigherOrSamePrec(operatorStack.peek(), token.charAt(0))) {
                    char operator = operatorStack.pop();
                    int b = operandStack.pop(), a =operandStack.pop();
                    int res = getResult(a,b,operator);
                    operandStack.push(res);
                }
                operatorStack.push(token.charAt(0));
            }
        }

        while(!operatorStack.empty()) {
            char operator = operatorStack.pop();
            int b = operandStack.pop(), a =operandStack.pop();
            int res = getResult(a,b,operator);
            operandStack.push(res);
        }
        // operand stack should have just 1 value, that will be the result
        return operandStack.pop();
    }

    private static boolean hasHigherOrSamePrec(Character peek, char token) {
        if(peek == '(' || peek == ')') return false;
        if(peek==token) return true;
        if(token == '/' ) return false;
        if(token == '*' && peek =='/') return true;
        return token == '+' || token == '-';

    }

    private static boolean isOperator(String token) {
        return "+*/-".contains(token);
    }

    private static int getResult(int a, int b, char operatorTop) throws DivideByZero {
        switch (operatorTop) {
            case '-': return a-b;
            case '+': return a+b;
            case '*': return a*b;
            case '/':
                if(b==0) throw new DivideByZero();
                return a/b;
            default:
                throw new IllegalArgumentException(""+operatorTop);
        }
    }

    private static boolean isOperand(String token) {
        return !DELIMS.contains(token);
    }
}
